"""Enable deterministic training for PyTorch models.

Note that there will be a performance penalty for enabling this feature.
"""
import os
import torch
from .plugin_base import PluginBase


class Plugin(PluginBase):
    def __init__(self):
        self._original_cudnn_benchmark = torch.backends.cudnn.benchmark
        self._original_cudnn_deterministic = torch.backends.cudnn.deterministic
        self._original_cublas_workspace_config = os.getenv('CUBLAS_WORKSPACE_CONFIG')

    def on_train_start(self, trainer, model):
        self._enable_deterministic()

    def on_train_end(self, trainer, model):
        self._disable_deterministic()

    def on_prediction_start(self, trainer, model):
        self._enable_deterministic()

    def on_prediction_end(self, trainer, model):
        self._disable_deterministic()

    def _enable_deterministic(self):
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
        torch.use_deterministic_algorithms(True, warn_only=True)
        os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8'

    def _disable_deterministic(self):
        torch.backends.cudnn.benchmark = self._original_cudnn_benchmark
        torch.backends.cudnn.deterministic = self._original_cudnn_deterministic
        torch.use_deterministic_algorithms(False)
        if self._original_cublas_workspace_config is not None:
            os.environ['CUBLAS_WORKSPACE_CONFIG'] = self._original_cublas_workspace_config
        else:
            del os.environ['CUBLAS_WORKSPACE_CONFIG']



import dataclasses
import logging
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Flatten a captioning dataset with multiple targets per image into a dataset with a single target per image.

    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    def execute(self, inputs):
        dataset = FlatDataset(inputs.dataset)
        return self.Outputs(dataset=dataset)

    def dry_run(self, inputs):
        return self.execute(inputs)


class FlatDataset(torch.utils.data.Dataset):
    def __init__(self, dataset):
        self._dataset = dataset

        if hasattr(dataset, 'get_targets'):
            num_targets = [len(dataset.get_targets(i)) for i in range(len(dataset))]
        else:
            num_targets = [len(dataset[i][1]) for i in range(len(dataset))]

        index_mapping = []
        for i, num_target in enumerate(num_targets):
            index_mapping.extend([(i, j) for j in range(num_target)])

        self._index_mapping = index_mapping

        logger.info(f"Flattened dataset from {len(dataset)} to {len(index_mapping)}")

    def __len__(self):
        return len(self._index_mapping)

    def __getitem__(self, index):
        dataset_index, target_index = self._index_mapping[index]
        d = self._dataset[dataset_index]
        assert isinstance(d[1], list)
        return d[0], [d[1][target_index]]

    def get_targets(self, index):
        dataset_index, target_index = self._index_mapping[index]
        if hasattr(self._dataset, 'get_targets'):
            targets = self._dataset.get_targets(dataset_index)
        else:
            targets = self._dataset[dataset_index][1]
        return [targets[target_index]]



import unittest
import torch
from irisml.tasks.evaluate_phrase_grounding import Task


class TestEvaluatePhraseGrounding(unittest.TestCase):
    def test_simple(self):
        predictions = [[((0, 1), torch.tensor([[0.1, 0.1, 0.2, 0.2]]))]]
        targets = [[((0, 1), torch.tensor([[0.1, 0.1, 0.2, 0.2]]))]]
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions=predictions, targets=targets))
        self.assertEqual(outputs.precision, 1.0)
        self.assertEqual(outputs.recall, 1.0)

        predictions = [[((0, 1), torch.tensor([[0.1, 0.1, 0.2, 0.2], [0, 0, 1.0, 1.0]]))]]
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions=predictions, targets=targets))
        self.assertEqual(outputs.precision, 1.0)
        self.assertEqual(outputs.recall, 1.0)

        predictions = [[((0, 1), torch.tensor([[0.8, 0.8, 0.9, 0.9], [0, 0, 1.0, 1.0]]))]]
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions=predictions, targets=targets))
        self.assertEqual(outputs.precision, 0.0)
        self.assertEqual(outputs.recall, 0.0)

        predictions = [[((0, 1), torch.tensor([[0.1, 0.1, 0.2, 0.2]])), ((1, 2), torch.tensor([[0.1, 0.1, 0.2, 0.2]]))]]
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions=predictions, targets=targets))
        self.assertEqual(outputs.precision, 1.0)  # Predictions with wrong text span are ignored.
        self.assertEqual(outputs.recall, 1.0)

        # Tuple might be converted to a list. This happens when CUDA is used.
        predictions = [[[[0, 1], torch.tensor([[0.1, 0.1, 0.2, 0.2]])], ([1, 2], torch.tensor([[0.1, 0.1, 0.2, 0.2]]))]]
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions=predictions, targets=targets))
        self.assertEqual(outputs.precision, 1.0)  # Predictions with wrong text span are ignored.
        self.assertEqual(outputs.recall, 1.0)

        predictions = [[]]
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions=predictions, targets=targets))
        self.assertEqual(outputs.precision, 0.0)  # No predictions, return zero.
        self.assertEqual(outputs.recall, 0.0)

        predictions = [[((0, 1), torch.zeros(0, 4))]]
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions=predictions, targets=targets))
        self.assertEqual(outputs.precision, 0.0)
        self.assertEqual(outputs.recall, 0.0)



import unittest
import PIL.Image
import torch.utils.data
from irisml.tasks.flatten_captioning_dataset import Task


class FakeDataset(torch.utils.data.Dataset):
    def __init__(self, data):
        self._data = data

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self):
        return len(self._data)


class TestFlattenCaptioningDataset(unittest.TestCase):
    def test_simple(self):
        dataset = FakeDataset([(PIL.Image.new('RGB', (8, 8)), ['caption A', 'caption B', 'caption C']), (PIL.Image.new('RGB', (16, 16)), []), (PIL.Image.new('RGB', (32, 32)), ['caption D'])])
        new_dataset = Task(Task.Config()).execute(Task.Inputs(dataset)).dataset
        self.assertEqual(len(new_dataset), 4)

        self.assertIsInstance(new_dataset[0][0], PIL.Image.Image)
        self.assertEqual(new_dataset[0][0].size, (8, 8))
        self.assertEqual(new_dataset[1][0].size, (8, 8))
        self.assertEqual(new_dataset[2][0].size, (8, 8))
        self.assertEqual(new_dataset[3][0].size, (32, 32))

        self.assertEqual(new_dataset[0][1], ['caption A'])
        self.assertEqual(new_dataset[1][1], ['caption B'])
        self.assertEqual(new_dataset[2][1], ['caption C'])
        self.assertEqual(new_dataset[3][1], ['caption D'])



import base64
import json
import pathlib
import tempfile
import unittest
import PIL.Image
import torch.utils.data
from irisml.tasks.save_jsonl_vqa_dataset import Task


class FakeDataset(torch.utils.data.Dataset):
    def __init__(self, data):
        super().__init__()
        self._data = data

    def __len__(self):
        return len(self._data)

    def __getitem__(self, index):
        return self._data[index]


class TestSaveJsonlVqaDataset(unittest.TestCase):
    def test_simple(self):
        dataset = FakeDataset([
            (('question1', PIL.Image.new('RGB', (32, 32))), 'answer1'),
            (('question2', PIL.Image.new('RGB', (32, 32))), 'answer2')
            ])

        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = pathlib.Path(temp_dir) / 'new_dir' / 'test.jsonl'
            outputs = Task(Task.Config(temp_file)).execute(Task.Inputs(dataset=dataset))
            self.assertIsNotNone(outputs)

            saved = [json.loads(line) for line in temp_file.read_text().splitlines()]
            self.assertEqual(len(saved), 2)
            self.assertEqual(saved[0]['id'], 1)
            self.assertEqual(saved[0]['question'], 'question1')
            self.assertEqual(saved[0]['answer'], 'answer1')
            self.assertTrue(base64.b64decode(saved[0]['image']))

            self.assertEqual(saved[1]['id'], 2)
            self.assertEqual(saved[1]['question'], 'question2')
            self.assertEqual(saved[1]['answer'], 'answer2')
            self.assertTrue(base64.b64decode(saved[0]['image']))



"""Plugin to record CUDA memory snapshots.

See https://pytorch.org/docs/stable/torch_cuda_memory.html for the details.
"""
import logging
import pathlib
import torch.cuda.memory
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, iter_count=5, path='outputs/memory_snapshot.pkl'):
        self._path = pathlib.Path(path)
        self._iter_count = iter_count
        self._done = False
        self._recording = False

    def on_train_start(self, trainer, model):
        if torch.distributed.is_initialized() and torch.distributed.get_rank() != 0:
            logger.debug(f"Skipping memory recording on rank {torch.distributed.get_rank()}")
            return

        torch.cuda.memory._record_memory_history()
        logger.info("Recording CUDA memory history")
        self._recording = True

    def on_train_end(self, trainer, model):
        torch.cuda.memory._record_memory_history(enabled=None)

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        if self._recording and batch_index == self._iter_count:
            torch.cuda.memory._dump_snapshot(self._path)
            logger.info(f"Memory snapshot saved to {self._path}")
            # Disable memory recording since it consumes significant amount of CPU memory
            torch.cuda.memory._record_memory_history(enabled=None)
            self._recording = False



import dataclasses
import logging
import typing
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Calculate accuracy of string matching.

    Case insensitive.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        predictions: typing.List[str]
        targets: typing.List[str]

    @dataclasses.dataclass
    class Outputs:
        accuracy: float

    def execute(self, inputs):
        if len(inputs.predictions) != len(inputs.targets):
            raise ValueError('predictions and targets must have same length')

        num_correct = 0
        for prediction, target in zip(inputs.predictions, inputs.targets):
            # Ignore cases
            if prediction.lower() == target.lower():
                num_correct += 1

        accuracy = num_correct / len(inputs.predictions)
        logger.info(f"Accuracy: {accuracy}, num_correct: {num_correct}, num_samples: {len(inputs.predictions)}")
        return self.Outputs(accuracy=accuracy)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
import PIL.Image
import torch
from irisml.tasks.convert_vqa_dataset_to_image_text_classification_dataset import Task


class TestConvertVqaDatasetToImageTextClassificationDataset(unittest.TestCase):
    def test_simple(self):
        dataset = [(('question1', PIL.Image.new('RGB', (10, 10))), 'answer1'), (('question2', PIL.Image.new('RGB', (10, 10))), 'answer2'), (('question3', PIL.Image.new('RGB', (10, 10))), 'answer1')]

        outputs = Task(Task.Config()).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs.dataset), 3)
        self.assertEqual(outputs.num_classes, 2)
        self.assertEqual(outputs.class_names, ['answer1', 'answer2'])
        self.assertIsInstance(outputs.dataset[0][0][0], PIL.Image.Image)
        self.assertEqual(outputs.dataset[0][0][1], 'question1')
        self.assertIsInstance(outputs.dataset[0][1], torch.Tensor)

        self.assertEqual(outputs.dataset[0][1], torch.tensor(0, dtype=torch.long))
        self.assertEqual(outputs.dataset[1][1], torch.tensor(1, dtype=torch.long))
        self.assertEqual(outputs.dataset[2][1], torch.tensor(0, dtype=torch.long))

    def test_invalid_dataset(self):
        dataset = [(PIL.Image.new('RGB', (8, 8)), 'answer1')]
        with self.assertRaises(ValueError):
            Task(Task.Config()).execute(Task.Inputs(dataset))

        dataset = [('question1', PIL.Image.new('RGB', (8, 8)), torch.tensor(0))]
        with self.assertRaises(ValueError):
            Task(Task.Config()).execute(Task.Inputs(dataset))

        dataset = []
        Task(Task.Config()).execute(Task.Inputs(dataset))  # Should not raise

    def test_class_names_input(self):
        dataset = [(('question1', PIL.Image.new('RGB', (10, 10))), 'answer1'), (('question2', PIL.Image.new('RGB', (10, 10))), 'answer2'), (('question3', PIL.Image.new('RGB', (10, 10))), 'answer1')]

        outputs = Task(Task.Config()).execute(Task.Inputs(dataset, class_names=['answer2', 'answer1', 'answer3']))
        self.assertEqual(len(outputs.dataset), 3)
        self.assertEqual(outputs.num_classes, 3)
        self.assertEqual(outputs.class_names, ['answer2', 'answer1', 'answer3'])

        self.assertEqual(outputs.dataset[0][1], torch.tensor(1, dtype=torch.long))
        self.assertEqual(outputs.dataset[1][1], torch.tensor(0, dtype=torch.long))
        self.assertEqual(outputs.dataset[2][1], torch.tensor(1, dtype=torch.long))

        # Should raise if class_names is not a subset of the dataset
        with self.assertRaises(ValueError):
            Task(Task.Config()).execute(Task.Inputs(dataset, class_names=['answer1']))  # missing 'answer2'



"""Pluging to profile model training and inference.

If you encounter the following error on WSL, visit https://github.com/pytorch/pytorch/issues/99615
"function cbapi->getCuptiStatus() failed with error CUPTI_ERROR_NOT_INITIALIZED"

Args:
    num_iters (int): Number of iterations to profile.
    trace_filename_prefix (str): Prefix for the trace filename. If None, the trace will not be saved.
"""

import datetime
import logging
import pathlib
import torch
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, num_iters=10, trace_filename_prefix='trace'):
        self._num_iters = num_iters
        self._trace_filename_prefix = trace_filename_prefix
        self._profiler = None

    def on_train_start(self, trainer, model):
        self._start()

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        if self._profiler is not None and batch_index >= self._num_iters - 1:
            self._stop()

    def on_train_end(self, trainer, model):
        self._stop()

    def on_prediction_start(self, trainer, model):
        self._start()

    def on_prediction_batch_end(self, trainer, model, batch, batch_index):
        if self._profiler is not None and batch_index >= self._num_iters - 1:
            self._stop()

    def on_prediction_end(self, trainer, model):
        self._stop()

    def _start(self):
        logger.info(f"Starting profiling for {self._num_iters} iterations")
        assert self._profiler is None
        self._profiler = torch.profiler.profile(activities=[torch.profiler.ProfilerActivity.CPU, torch.profiler.ProfilerActivity.CUDA])
        self._profiler.__enter__()

    def _stop(self):
        if self._profiler is not None:
            self._profiler.__exit__(None, None, None)
            if self._trace_filename_prefix:
                filename = f'{self._trace_filename_prefix}_{datetime.datetime.now().strftime("%Y%m%d_%H%M%S")}.json'
                pathlib.Path(filename).parent.mkdir(parents=True, exist_ok=True)
                logger.info(f"Exporting trace to {filename}")
                self._profiler.export_chrome_trace(filename)
            self._profiler = None



import pickle
import unittest
import PIL.Image
import torch
import torchvision.transforms
from irisml.tasks.make_vqa_image_transform import Task


class TestMakeVqaImageTransform(unittest.TestCase):
    def test_simple_without_tokenizer(self):
        image_transform = torchvision.transforms.ToTensor()

        def text_transform(text):
            return f'question: {text} answer:'

        outputs = Task(Task.Config()).execute(Task.Inputs(image_transform, text_transform))
        transform = outputs.transform

        transform_outputs = transform(('What is this?', PIL.Image.new('RGB', (32, 32))), 'R2D2')
        self.assertEqual(transform_outputs[0][0], 'question: What is this? answer:')
        self.assertIsInstance(transform_outputs[0][1], torch.Tensor)
        self.assertEqual(transform_outputs[1], 'R2D2')

    def test_simple_with_tokenizer(self):
        def tokenizer(text):
            return torch.tensor([1, 2, 3]), torch.tensor([1, 1, 0])

        image_transform = torchvision.transforms.ToTensor()
        outputs = Task(Task.Config()).execute(Task.Inputs(image_transform, tokenizer=tokenizer))
        transform = outputs.transform

        transform_outputs = transform(('What is this?', PIL.Image.new('RGB', (32, 32))), 'R2D2')
        self.assertIsInstance(transform_outputs[0][0], tuple)
        self.assertEqual(len(transform_outputs[0][0]), 2)
        self.assertIsInstance(transform_outputs[0][0][0], torch.Tensor)
        self.assertIsInstance(transform_outputs[0][0][1], torch.Tensor)
        self.assertIsInstance(transform_outputs[0][1], torch.Tensor)

        self.assertIsInstance(transform_outputs[1], tuple)
        self.assertEqual(len(transform_outputs[1]), 2)
        self.assertIsInstance(transform_outputs[1][0], torch.Tensor)
        self.assertIsInstance(transform_outputs[1][1], torch.Tensor)

    def test_collate_function(self):
        image_transform = torchvision.transforms.ToTensor()
        outputs = Task(Task.Config()).execute(Task.Inputs(image_transform))
        collate_function = outputs.collate_function

        pickle.dumps(collate_function)  # Make sure it's pickleable

        question = (torch.Tensor([1, 2, 3]), torch.Tensor([1, 1, 0]))
        answer = (torch.Tensor([4, 5, 6]), torch.Tensor([1, 1, 0]))
        image = torch.rand(3, 32, 32)

        collate_outputs = collate_function([((question, image), answer) for _ in range(4)])
        self.assertIsInstance(collate_outputs, tuple)
        self.assertEqual(len(collate_outputs), 2)

        # inputs
        self.assertIsInstance(collate_outputs[0], tuple)
        self.assertEqual(len(collate_outputs[0]), 2)
        self.assertIsInstance(collate_outputs[0][0], tuple)
        self.assertEqual(len(collate_outputs[0][0]), 2)
        self.assertIsInstance(collate_outputs[0][0][0], torch.Tensor)
        self.assertIsInstance(collate_outputs[0][0][1], torch.Tensor)
        self.assertIsInstance(collate_outputs[0][1], torch.Tensor)
        self.assertEqual(collate_outputs[0][0][0].shape, (4, 3))
        self.assertEqual(collate_outputs[0][0][1].shape, (4, 3))
        self.assertEqual(collate_outputs[0][1].shape, (4, 3, 32, 32))

        # targets
        self.assertIsInstance(collate_outputs[1], tuple)
        self.assertEqual(len(collate_outputs[1]), 2)
        self.assertIsInstance(collate_outputs[1][0], torch.Tensor)
        self.assertIsInstance(collate_outputs[1][1], torch.Tensor)
        self.assertEqual(collate_outputs[1][0].shape, (4, 3))
        self.assertEqual(collate_outputs[1][1].shape, (4, 3))

        # When targets are strings
        collate_outputs = collate_function([((question, image), 'answer') for _ in range(4)])
        self.assertIsInstance(collate_outputs[1], list)
        self.assertEqual(len(collate_outputs[1]), 4)
        self.assertIsInstance(collate_outputs[1][0], str)
        self.assertEqual(collate_outputs[1], ['answer' for _ in range(4)])

    def test_collate_function_padding(self):
        image_transform = torchvision.transforms.ToTensor()
        outputs = Task(Task.Config(pad_token_id=100)).execute(Task.Inputs(image_transform))
        collate_function = outputs.collate_function

        collate_outputs = collate_function([(((torch.Tensor([i] * i), torch.Tensor([1] * i)), torch.rand(3, 32, 32)), 'answer') for i in range(1, 5)])

        self.assertEqual(collate_outputs[0][0][0].tolist(), [[1, 100, 100, 100], [2, 2, 100, 100], [3, 3, 3, 100], [4, 4, 4, 4]])
        self.assertEqual(collate_outputs[0][0][1].tolist(), [[1, 0, 0, 0], [1, 1, 0, 0], [1, 1, 1, 0], [1, 1, 1, 1]])
        self.assertEqual(collate_outputs[1], ['answer' for _ in range(4)])



import unittest
import PIL.Image
import torch
import torchvision.transforms
from irisml.tasks.make_phrase_grounding_image_transform import Task


class TestMakePhraseGroundingImageTransform(unittest.TestCase):
    def test_simple(self):
        image_transform = torchvision.transforms.ToTensor()

        def text_transform(x):
            return x + '!'

        transform = Task(Task.Config()).execute(Task.Inputs(image_transform, text_transform)).transform

        outputs = transform(('caption', PIL.Image.new('RGB', (100, 100))), [((0, 1), torch.tensor([[0, 0, 1, 1]]))])
        self.assertEqual(outputs[0][0], 'caption!')
        self.assertIsInstance(outputs[0][1], torch.Tensor)
        self.assertEqual(outputs[1][0][0], (0, 1))
        self.assertEqual(outputs[1][0][1].tolist(), [[0, 0, 1, 1]])



import unittest
from irisml.tasks.evaluate_string_matching_accuracy import Task


class TestEvaluateStringMatchingAccuracy(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config()).execute(Task.Inputs(['A', 'b', 'c'], ['a', 'B', 'C']))
        self.assertEqual(outputs.accuracy, 1.0)

        outputs = Task(Task.Config()).execute(Task.Inputs(['aa', 'bb', 'cc'], ['a', 'b', 'c']))
        self.assertEqual(outputs.accuracy, 0.0)

        outputs = Task(Task.Config()).execute(Task.Inputs(['aa', 'bb'], ['aa', 'bc']))
        self.assertEqual(outputs.accuracy, 0.5)



import base64
import dataclasses
import io
import json
import logging
import pathlib
import pickle
import tempfile
import typing
import PIL.Image
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Load a VQA dataset from a jsonl file.

    JSONL dataset format: {image: <base64 encoded image>, question: str, answer: str}

    Loaded dataset format is a tuple of ((question: str, image: PIL.Image.Image), answer: str)

    The dataset is cached in a temporary directory.

    Config:
        filepath (Path): Path to the jsonl file.
        cache_dirpath (Optional[Path]): Path to the directory where the dataset will be cached. If not specified, a temporary directory will be created.
            Caller is responsible for deleting the directory when it is no longer needed.
    """
    VERSION = '0.2.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Config:
        filepath: pathlib.Path
        cache_dirpath: typing.Optional[pathlib.Path] = None

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    def execute(self, inputs):
        dataset = VQADataset(self.config.filepath, self.config.cache_dirpath)
        return self.Outputs(dataset=dataset)

    def dry_run(self, inputs):
        return self.execute(inputs)


class VQADataset(torch.utils.data.Dataset):
    def __init__(self, filepath, cache_dirpath):
        super().__init__()
        if cache_dirpath:
            self._temp_dir = None
            self._temp_dir_path = cache_dirpath
            self._temp_dir_path.mkdir(parents=True, exist_ok=True)
        else:
            self._temp_dir = tempfile.TemporaryDirectory()
            self._temp_dir_path = pathlib.Path(self._temp_dir.name)

        logger.info(f"Unpacking {filepath} to {self._temp_dir_path}")
        index = 0
        with open(filepath) as f:
            for line in f:
                if not line:
                    continue
                json_data = json.loads(line)
                data = (base64.b64decode(json_data['image']), json_data['question'], json_data['answer'])
                (self._temp_dir_path / f'{index}.pkl').write_bytes(pickle.dumps(data))
                index += 1
        logger.info(f"Loaded {index} images from {filepath}")
        self._num_images = index

    def __len__(self):
        return self._num_images

    def __getitem__(self, index):
        if index >= self._num_images:
            raise IndexError(f"Index {index} out of range")

        if not self._temp_dir_path:
            raise RuntimeError("Dataset is closed")

        image_bytes, question, answer = pickle.loads((self._temp_dir_path / f'{index}.pkl').read_bytes())
        image = PIL.Image.open(io.BytesIO(image_bytes))
        return (question, image), answer

    def get_targets(self, index):
        if index >= self._num_images:
            raise IndexError(f"Index {index} out of range")

        if not self._temp_dir_path:
            raise RuntimeError("Dataset is closed")

        _, _, answer = pickle.loads((self._temp_dir_path / f'{index}.pkl').read_bytes())
        return answer



import base64
import io
import pathlib
import tempfile
import unittest
import PIL.Image
from irisml.tasks.load_jsonl_vqa_dataset import Task


class TestLoadJsonlVqaDataset(unittest.TestCase):
    def test_load_jsonl_vqa_dataset(self):
        with tempfile.NamedTemporaryFile() as f, tempfile.TemporaryDirectory() as temp_dir:
            temp_file = pathlib.Path(f.name)
            base64_image = self._get_fake_base64_image()
            temp_file.write_text(f'{{"image": "{base64_image}", "question": "What is this?", "answer": "a cat"}}\n{{"image": "{base64_image}", "question": "What is that?", "answer": "a dog"}}')

            outputs = Task(Task.Config(temp_file, pathlib.Path(temp_dir))).execute(Task.Inputs())
            self.assertEqual(len(outputs.dataset), 2)
            self.assertIsInstance(outputs.dataset[0], tuple)
            self.assertIsInstance(outputs.dataset[0][0][0], str)
            self.assertIsInstance(outputs.dataset[0][0][1], PIL.Image.Image)
            self.assertIsInstance(outputs.dataset[0][1], str)
            self.assertEqual(outputs.dataset[0][1], 'a cat')

            self.assertEqual(outputs.dataset.get_targets(1), 'a dog')

    @staticmethod
    def _get_fake_base64_image():
        image = PIL.Image.new('RGB', (32, 32))
        with io.BytesIO() as f:
            image.save(f, format='PNG')
            return base64.b64encode(f.getvalue()).decode('utf-8')



import pickle
import unittest
import PIL.Image
import torch
from irisml.tasks.make_vqa_collate_function import Task


class TestMakeVqaCollateFunction(unittest.TestCase):
    def test_collate_function(self):
        outputs = Task(Task.Config()).execute(Task.Inputs())
        collate_function = outputs.collate_function

        pickle.dumps(collate_function)  # Make sure it's pickleable

        question = (torch.Tensor([1, 2, 3]), torch.Tensor([1, 1, 0]))
        answer = (torch.Tensor([4, 5, 6]), torch.Tensor([1, 1, 0]))
        image = torch.rand(3, 32, 32)

        collate_outputs = collate_function([((question, image), answer) for _ in range(4)])
        self.assertIsInstance(collate_outputs, tuple)
        self.assertEqual(len(collate_outputs), 2)

        # inputs
        self.assertIsInstance(collate_outputs[0], tuple)
        self.assertEqual(len(collate_outputs[0]), 2)
        self.assertIsInstance(collate_outputs[0][0], tuple)
        self.assertEqual(len(collate_outputs[0][0]), 2)
        self.assertIsInstance(collate_outputs[0][0][0], torch.Tensor)
        self.assertIsInstance(collate_outputs[0][0][1], torch.Tensor)
        self.assertIsInstance(collate_outputs[0][1], torch.Tensor)
        self.assertEqual(collate_outputs[0][0][0].shape, (4, 3))
        self.assertEqual(collate_outputs[0][0][1].shape, (4, 3))
        self.assertEqual(collate_outputs[0][1].shape, (4, 3, 32, 32))

        # targets
        self.assertIsInstance(collate_outputs[1], tuple)
        self.assertEqual(len(collate_outputs[1]), 2)
        self.assertIsInstance(collate_outputs[1][0], torch.Tensor)
        self.assertIsInstance(collate_outputs[1][1], torch.Tensor)
        self.assertEqual(collate_outputs[1][0].shape, (4, 3))
        self.assertEqual(collate_outputs[1][1].shape, (4, 3))

        # When targets are strings
        collate_outputs = collate_function([((question, image), 'answer') for _ in range(4)])
        self.assertIsInstance(collate_outputs[1], list)
        self.assertEqual(len(collate_outputs[1]), 4)
        self.assertIsInstance(collate_outputs[1][0], str)
        self.assertEqual(collate_outputs[1], ['answer' for _ in range(4)])

        # Phrase grounding with GPT-V
        collate_outputs = collate_function([(('caption', PIL.Image.new('RGB', (32, 32))), [((0, 2), torch.tensor([[0.1, 0.1, 0.2, 0.2]]))]) for _ in range(4)])
        self.assertEqual(collate_outputs[0][0], ['caption'] * 4)
        self.assertEqual(collate_outputs[0][1], [PIL.Image.new('RGB', (32, 32))] * 4)
        self.assertIsInstance(collate_outputs[1], list)
        self.assertEqual(len(collate_outputs[1]), 4)

        self.assertEqual(collate_outputs[1][0][0][0], (0, 2))

        # image_text_classification. Scalar LongTensor target.
        collate_outputs = collate_function([((question, image), torch.tensor(i)) for i in range(4)])
        self.assertIsInstance(collate_outputs[1], torch.Tensor)
        self.assertEqual(collate_outputs[1].tolist(), [0, 1, 2, 3])

    def test_collate_function_padding(self):
        outputs = Task(Task.Config(pad_token_id=100)).execute(Task.Inputs())
        collate_function = outputs.collate_function

        collate_outputs = collate_function([(((torch.Tensor([i] * i), torch.Tensor([1] * i)), torch.rand(3, 32, 32)), 'answer') for i in range(1, 5)])

        self.assertEqual(collate_outputs[0][0][0].tolist(), [[1, 100, 100, 100], [2, 2, 100, 100], [3, 3, 3, 100], [4, 4, 4, 4]])
        self.assertEqual(collate_outputs[0][0][1].tolist(), [[1, 0, 0, 0], [1, 1, 0, 0], [1, 1, 1, 0], [1, 1, 1, 1]])
        self.assertEqual(collate_outputs[1], ['answer' for _ in range(4)])



import collections
import dataclasses
import logging
import math
import string
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Evaluate captioning prediction results.

    Supported metrics are BLEU-1, BLEU-2, BLEU-3, BLEU-4 and CIDEr.

    Note that this task uses the simplest tokenizer. It might have different results from the pycocoevalcap results.

    This task is experimental. The evaluation result might not be accurate.
    """
    VERSION = '0.1.1'

    CIDER_SIGMA = 6.0

    @dataclasses.dataclass
    class Inputs:
        predictions: typing.List[str]
        targets: typing.List[typing.List[str]]

    @dataclasses.dataclass
    class Outputs:
        bleu1: float
        bleu2: float
        bleu3: float
        bleu4: float
        cider: float

    def execute(self, inputs):
        if len(inputs.predictions) != len(inputs.targets):
            raise ValueError(f"Number of predictions ({len(inputs.predictions)}) does not match number of targets ({len(inputs.targets)})")

        if len(inputs.predictions) == 0:
            raise ValueError("No predictions provided")

        logger.info(f"Evaluating {len(inputs.predictions)} predictions")

        total_pred_ngrams = collections.Counter()
        document_frequency = collections.Counter()

        all_pred_ngrams = []
        all_target_ngrams = []

        bleu_numerator = [0, 0, 0, 0]
        bleu_denominator = [0, 0, 0, 0]
        bleu_preds_length = 0
        bleu_targets_length = 0
        remove_punctuation = str.maketrans('', '', string.punctuation)
        for pred, targets in zip(inputs.predictions, inputs.targets):
            tokenized_pred = pred.translate(remove_punctuation).lower().split()
            tokenized_targets = [t.translate(remove_punctuation).lower().split() for t in targets]

            bleu_preds_length += len(tokenized_pred)
            bleu_targets_length += len(min(tokenized_targets, key=lambda x: abs(len(tokenized_pred) - len(x))))  # Closest length to prediction

            # Get the count of N grams
            pred_ngrams = self._get_ngrams(tokenized_pred)
            target_ngrams = [self._get_ngrams(t) for t in tokenized_targets]
            total_target_ngrams_for_image = collections.Counter()
            for target_ngram in target_ngrams:
                total_target_ngrams_for_image |= target_ngram

            # Save total counts for CIDEr
            total_pred_ngrams |= pred_ngrams
            document_frequency.update(set(total_target_ngrams_for_image))
            all_pred_ngrams.append(pred_ngrams)
            all_target_ngrams.append(target_ngrams)

            # Count the numbers for BLEU score
            intersection = pred_ngrams & total_target_ngrams_for_image
            for key, value in intersection.items():
                bleu_numerator[len(key) - 1] += value
            for key, value in pred_ngrams.items():
                bleu_denominator[len(key) - 1] += value

        # Calculate BLEU score
        bleu_scores = []
        bleu = 1.0
        penalty = 1.0 if bleu_preds_length > bleu_targets_length else math.exp(1 - bleu_targets_length / bleu_preds_length)
        for i in range(4):
            bleu *= (bleu_numerator[i] + 1e-15) / (bleu_denominator[i] + 1e-9)  # Add epsilon to avoid zero division and zero bleu.
            bleu_scores.append((bleu ** (1 / (i + 1))) * penalty)

        # Calculate CIDEr score
        log_num_targets = math.log(len(inputs.targets))
        cider_scores = []
        for pred_ngrams, target_ngrams in zip(all_pred_ngrams, all_target_ngrams):
            pred_vector = self._get_cider_vec(pred_ngrams, document_frequency, log_num_targets)
            pred_length = self._get_length_from_ngram(pred_ngrams)
            scores = torch.zeros((4,))
            for t in target_ngrams:
                target_vector = self._get_cider_vec(t, document_frequency, log_num_targets)
                target_length = self._get_length_from_ngram(t)
                scores += self._calculate_cider_similarity(pred_vector, target_vector, pred_length, target_length)
            score = scores.mean() / len(target_ngrams) * 10.0
            cider_scores.append(score)
        cider_score = sum(cider_scores) / len(cider_scores)

        logger.info(f"BLEU-1: {bleu_scores[0]:.4f}, BLEU-2: {bleu_scores[1]:.4f}, BLEU-3: {bleu_scores[2]:.4f}, BLEU-4: {bleu_scores[3]:.4f}, CIDEr: {cider_score:.4f}")
        return self.Outputs(bleu1=float(bleu_scores[0]), bleu2=float(bleu_scores[1]), bleu3=float(bleu_scores[2]), bleu4=float(bleu_scores[3]), cider=float(cider_score))

    def dry_run(self, inputs):
        return self.execute(inputs)

    @staticmethod
    def _get_ngrams(tokens):
        ngrams = collections.Counter()
        for n in range(1, 5):
            for i in range(len(tokens) - n + 1):
                ngrams[tuple(tokens[i:i + n])] += 1
        return ngrams

    @staticmethod
    def _get_length_from_ngram(ngram):
        return sum([value for key, value in ngram.items() if len(key) == 1])

    @staticmethod
    def _get_cider_vec(ngrams, document_ngrams, log_num_targets):
        vec = [collections.defaultdict(float) for _ in range(4)]
        for ngram, count in ngrams.items():
            df = math.log(max(1.0, document_ngrams[ngram]))
            vec[len(ngram) - 1][ngram] = count * (log_num_targets - df)
        return vec

    def _calculate_cider_similarity(self, pred_vec, target_vec, pred_length, target_length):
        scores = torch.zeros((4,))
        delta = pred_length - target_length

        for i in range(4):
            for ngram, count in pred_vec[i].items():
                scores[i] += min(count, target_vec[i][ngram]) * target_vec[i][ngram]

            pred_norm = math.sqrt(sum(count * count for count in pred_vec[i].values()))
            target_norm = math.sqrt(sum(count * count for count in target_vec[i].values()))

            if pred_norm != 0 and target_norm != 0:
                scores[i] /= pred_norm * target_norm

            scores[i] *= math.exp(-delta ** 2 / (2 * (self.CIDER_SIGMA ** 2)))

        return scores



import dataclasses
from typing import Callable, Optional, Tuple
import PIL.Image
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Creates a transform function for VQA task.

    The input to the transform function is ((question, image), targets), where question is a string,
    image is a PIL image, and targets is a string.

    If tokenizer is provided, the output of the transform function is ((question, image_tensor), targets),
    where question is a tuple of tensors, image_tensor is a tensor, and targets is a tuple of tensors.

    If tokenizer is not provided, the output of the transform function is ((question, image_tensor), targets),
    where question is a string, image_tensor is a tensor, and targets is a string.

    Config:
        skip_traget_tokenization (bool): If True, the target is not tokenized. Default: False
        pad_token_id (int): The id of the padding token. Used in the collate_function. Default: 1
    """
    VERSION = '1.2.0'

    @dataclasses.dataclass
    class Inputs:
        image_transform: Callable[[PIL.Image.Image], torch.Tensor]
        text_transform: Optional[Callable[[str], str]] = None
        target_text_transform: Optional[Callable[[str], str]] = None
        tokenizer: Optional[Callable[[str], Tuple[torch.Tensor, torch.Tensor]]] = None  # The output is (input_ids, attention_mask)

    @dataclasses.dataclass
    class Config:
        skip_target_tokenization: bool = False
        pad_token_id: int = 1

    @dataclasses.dataclass
    class Outputs:
        transform: Callable[[Tuple[str, PIL.Image.Image], str], Tuple[Tuple[str, torch.Tensor], str]]
        collate_function: Callable  # [(((input_ids, attention_mask), image_tensor), (input_ids, attention_mask))] => (((input_ids, attention_mask), image_tensor), (input_ids, attention_mask))

    def execute(self, inputs):
        transform = VqaImageTransform(inputs.image_transform, inputs.text_transform, inputs.target_text_transform, inputs.tokenizer, self.config.skip_target_tokenization)
        collate_function = Collate(self.config.pad_token_id)
        return self.Outputs(transform=transform, collate_function=collate_function)

    def dry_run(self, inputs):
        return self.execute(inputs)


class Collate:
    def __init__(self, pad_token_id):
        self._pad_token_id = pad_token_id

    def __call__(self, batch):
        questions = [b[0][0] for b in batch]
        images = [b[0][1] for b in batch]
        targets = [b[1] for b in batch]

        # input_ids, attention_mask. Use 0 for attention_mask padding.
        questions = (self._padded_stack([q[0] for q in questions], self._pad_token_id), self._padded_stack([q[1] for q in questions]))
        images = torch.stack(images, 0)
        if not isinstance(targets[0], str):
            targets = (self._padded_stack([t[0] for t in targets], self._pad_token_id), self._padded_stack([t[1] for t in targets]))

        return (questions, images), targets

    def _padded_stack(self, tensors, pad_token_id=0):
        """Stacks tensors along the first dimension and pads them to the same length."""
        max_length = max([t.shape[0] for t in tensors])
        return torch.stack([torch.nn.functional.pad(t, (0, max_length - t.shape[0]), value=pad_token_id) for t in tensors], 0)


class VqaImageTransform:
    def __init__(self, image_transform, text_transform, target_text_transform, tokenizer, skip_target_tokenization):
        self._image_transform = image_transform
        self._text_transform = text_transform
        self._target_text_transform = target_text_transform
        self._tokenizer = tokenizer
        self._skip_target_tokenization = skip_target_tokenization

    def __call__(self, inputs, targets):
        question, image = inputs
        assert isinstance(question, str)
        assert isinstance(image, PIL.Image.Image)
        assert isinstance(targets, str)

        image_tensor = self._image_transform(image)

        if self._text_transform:
            question = self._text_transform(question)

        if self._target_text_transform:
            targets = self._target_text_transform(targets)

        if self._tokenizer:
            question = self._tokenizer(question)
            if not self._skip_target_tokenization:
                targets = self._tokenizer(targets)

        return (question, image_tensor), targets



import dataclasses
import typing
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Creates a collate_function for Visual Question Answering (VQA) and Phrase Grounding task.

    Config:
        pad_token_id (int): The id of the padding token. Default: 1.
    """
    VERSION = '0.2.1'

    @dataclasses.dataclass
    class Config:
        pad_token_id: int = 1

    @dataclasses.dataclass
    class Outputs:
        collate_function: typing.Callable  # [(((input_ids, attention_mask), image_tensor), (input_ids, attention_mask))] => (((input_ids, attention_mask), image_tensor), (input_ids, attention_mask))

    def execute(self, inputs):
        collate_function = Collate(self.config.pad_token_id)
        return self.Outputs(collate_function)

    def dry_run(self, inputs):
        return self.execute(inputs)


class Collate:
    def __init__(self, pad_token_id):
        self._pad_token_id = pad_token_id

    def __call__(self, batch):
        questions = [b[0][0] for b in batch]
        images = [b[0][1] for b in batch]
        targets = [b[1] for b in batch]

        # input_ids, attention_mask. Use 0 for attention_mask padding.
        if isinstance(questions[0][0], torch.Tensor):
            questions = (self._padded_stack([q[0] for q in questions], self._pad_token_id), self._padded_stack([q[1] for q in questions]))
        if isinstance(images[0], torch.Tensor):
            images = torch.stack(images, 0)
        if isinstance(targets[0], torch.Tensor):
            targets = torch.stack(targets, 0)
        elif isinstance(targets[0][0], torch.Tensor):
            targets = (self._padded_stack([t[0] for t in targets], self._pad_token_id), self._padded_stack([t[1] for t in targets]))

        return (questions, images), targets

    def _padded_stack(self, tensors, pad_token_id=0):
        """Stacks tensors along the first dimension and pads them to the same length."""
        max_length = max([t.shape[0] for t in tensors])
        return torch.stack([torch.nn.functional.pad(t, (0, max_length - t.shape[0]), value=pad_token_id) for t in tensors], 0)



import unittest
from irisml.tasks.evaluate_captioning import Task


class TestEvaluateCaptioning(unittest.TestCase):
    def test_perfect(self):
        predictions = ['Hello World Good bye!!', 'This is perfect test']
        ground_truths = [['Hello World good Bye'], ['This is perfect test', 'This is perfect test']]
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions, ground_truths))
        self.assertAlmostEqual(outputs.bleu1, 1.0)
        self.assertAlmostEqual(outputs.bleu2, 1.0)
        self.assertAlmostEqual(outputs.bleu3, 1.0)
        self.assertAlmostEqual(outputs.bleu4, 1.0)
        self.assertAlmostEqual(outputs.cider, 10.0)
        self.assertIsInstance(outputs.bleu1, float)
        self.assertIsInstance(outputs.bleu2, float)
        self.assertIsInstance(outputs.bleu3, float)
        self.assertIsInstance(outputs.bleu4, float)
        self.assertIsInstance(outputs.cider, float)

    def test_totally_wrong(self):
        predictions = ['How are you today', 'Only wrong words here']
        ground_truths = [['Hello World good Bye'], ['This is perfect test', 'This is perfect test']]
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions, ground_truths))
        self.assertAlmostEqual(outputs.bleu1, 0.0)
        self.assertAlmostEqual(outputs.bleu2, 0.0)
        self.assertAlmostEqual(outputs.bleu3, 0.0)
        self.assertAlmostEqual(outputs.bleu4, 0.0)
        self.assertAlmostEqual(outputs.cider, 0.0)



"""Save state_dict of the training model during training.

Parameters:
    filename_prefix (str): Prefix of the filename to save the state_dict of the model.
    epoch_interval (int): Interval of epochs to save the state_dict of the model.
"""

import logging
import os
import torch
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, filename_prefix='model_snapshot_', epoch_interval=1):
        self._filename_prefix = filename_prefix
        self._epoch_interval = epoch_interval

    def on_train_epoch_end(self, trainer, model, epoch_index):
        if (epoch_index + 1) % self._epoch_interval == 0:
            filename = f'{self._filename_prefix}{epoch_index}.pth'
            logger.info(f"Saving state_dict of the model to {filename}")
            torch.save(model.state_dict(), filename)
            file_size = os.stat(filename).st_size
            logger.info(f"Saved state_dict of the model to {filename}. File size: {file_size} bytes.")



"""Log metrics on Azure ML.

This plugin requires mlflow and azureml-mlflow packages. Use "azureml" extra dependency to install them.
```
pip install irisml-tasks-training[azureml]
```

"""
import logging
import mlflow
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def on_train_start(self, trainer, model):
        logger.info("Enabling MLflow autologging")
        mlflow.autolog(log_models=False, exclusive=True)

    def on_train_end(self, trainer, model):
        logger.info("Disabling MLflow autologging")
        mlflow.autolog(disable=True)

    def on_train_backward_start(self, trainer, model, loss):
        # Get a loss in this method, because other plugins may change it.
        mlflow.log_metric("train_loss", float(loss.item()))
        return loss

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        for i, lr in enumerate(g['lr'] for g in trainer.optimizer.param_groups):
            mlflow.log_metric(f"lr{i}", float(lr))

    def on_validation_batch_end(self, trainer, model, loss, batch, batch_index):
        mlflow.log_metric("val_loss", float(loss.item()))



"""Plugin to debug training process.

Args:
    level (int): Log level. 0=info, 1=debug, 2=verbose. Default: 0.
"""
import logging
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, level=0):
        self._level = level
        self._batch_loss = 0

    def on_train_backward_start(self, trainer, model, loss):
        # Get a loss in this method, because other plugins may change it.
        self._batch_loss = float(loss.item())
        return loss

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        lrs = [g['lr'] for g in trainer.optimizer.param_groups]
        if batch_index % 100 == 0:
            self._log0(f"Batch {batch_index} loss: {self._batch_loss}, lr: {lrs}")
        else:
            self._log2(f"Batch {batch_index} loss: {self._batch_loss}, lr: {lrs}")

    def _log0(self, msg):
        if self._level >= 0:
            logger.info(msg)

    def _log1(self, msg):
        if self._level >= 1:
            logger.info(msg)

    def _log2(self, msg):
        if self._level >= 2:
            logger.info(msg)



import unittest
import pathlib
import tempfile
import json
from irisml.tasks.extract_val_by_key_from_jsonl import Task


class TestExtractValByKeyFromJsonl(unittest.TestCase):
    def test_extract_values(self):
        data = [{"a": 1, "b": 2},
                {"a": 3, "b": 4},
                {"a": 5, "b": 6}]

        with tempfile.TemporaryDirectory() as temp_dir:
            jsonl_file = pathlib.Path(temp_dir) / "test.jsonl"
            jsonl_file.write_text("\n".join([json.dumps(d) for d in data]))

            outputs = Task(Task.Config(key='a')).execute(Task.Inputs(jsonl_file_path=jsonl_file))
            self.assertEqual(outputs.results, [1, 3, 5])



import dataclasses
import logging
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Convert VQA dataset to image text classification dataset.

    This task consideres VQA question as text input and answer as class label.

    The expected VQA dataset format is (question, image), answer, where question is a string, image is a PIL image, and answer is a string.

    If class_names is given, it is used as class names. Otherwise, class names are extracted from the dataset.
    """
    VERSION = '0.2.0'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset
        class_names: typing.Optional[typing.List[str]] = None

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset
        num_classes: int
        class_names: typing.List[str]

    def execute(self, inputs):
        # Verify the dataset is VQA dataset by checking the first data.
        if inputs.dataset:
            first_data = inputs.dataset[0]
            if not isinstance(first_data, (tuple, list)):
                raise ValueError(f"Unexpected dataset type: {type(first_data)}")
            if len(first_data) != 2:
                raise ValueError(f"Unexpected dataset length: {len(first_data)}")
            if not isinstance(first_data[0], (tuple, list)):
                raise ValueError(f"This task expects (text, image) tuple as dataset input, but got {type(first_data[0])}")
            if not isinstance(first_data[1], str):
                raise ValueError(f"This task expects string as target, but got {type(first_data[1])}")
        else:
            logger.warning("Empty dataset is given.")

        class_names = self._get_class_names(inputs.dataset)
        logger.debug(f"Extracted class_names from dataset. num_classes={len(class_names)} class_names={class_names}")

        if inputs.class_names:
            missing_class_names = set(class_names) - set(inputs.class_names)
            if missing_class_names:
                raise ValueError(f"Given class_names is missing some class names: {missing_class_names}")
            class_names = inputs.class_names
            logger.info(f"Using given class_names. num_classes={len(class_names)} class_names={class_names}")

        dataset = Dataset(inputs.dataset, class_names)
        logger.info(f"Converted dataset. num_classes={len(class_names)} num_samples={len(dataset)}")
        return self.Outputs(dataset=dataset, num_classes=len(class_names), class_names=class_names)

    def dry_run(self, inputs):
        return self.execute(inputs)

    def _get_class_names(self, dataset):
        answer_set = set()
        if hasattr(dataset, 'get_targets'):
            for i in range(len(dataset)):
                answer_set.add(dataset.get_targets(i))
        else:
            for _, answer in dataset:
                answer_set.add(answer)

        return sorted(list(answer_set))


class Dataset(torch.utils.data.Dataset):
    def __init__(self, dataset, class_names):
        self._class_names = class_names
        self._class_name_to_index = {class_name: i for i, class_name in enumerate(class_names)}
        self._dataset = dataset

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, index):
        (question, image), targets = self._dataset[index]
        assert isinstance(targets, str)
        assert targets in self._class_name_to_index
        class_id = torch.tensor(self._class_name_to_index[targets], dtype=torch.long)
        return (image, question), class_id

    def get_targets(self, index):
        if hasattr(self._dataset, 'get_targets'):
            targets = self._dataset.get_targets(index)
        else:
            _, targets = self._dataset[index]

        class_id = torch.tensor(self._class_name_to_index[targets], dtype=torch.long)
        return class_id



import unittest
import PIL.Image
import torch
from irisml.tasks.concatenate_datasets import Task


from utils import FakeDataset, FakeDatasetWithGetTargets


class TestConcatenateDatasets(unittest.TestCase):
    def test_same_classsspace_classification(self):
        dataset0 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        dataset1 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        outputs = Task(Task.Config(use_same_classspace=True)).execute(Task.Inputs(dataset0, dataset1))
        self.assertEqual(len(outputs.dataset), 6)
        self.assertEqual(outputs.dataset[0], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[3], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[5], (None, torch.tensor(3)))

    def test_class_names_pass_through(self):
        dataset0 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        dataset1 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        class_names0 = ['label0', 'label1', 'label2', 'label3', 'label4']
        outputs = Task(Task.Config(use_same_classspace=True)).execute(Task.Inputs(dataset0, dataset1, class_names0))
        self.assertEqual(outputs.class_names, class_names0)
        self.assertEqual(outputs.num_classes, len(class_names0))

    def test_same_classspace_with_different_class_names(self):
        dataset0 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        dataset1 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        class_names0 = ['label0', 'label1', 'label2', 'label3', 'label4']
        class_names1 = ['label5', 'label1', 'label2', 'label3', 'label4']
        with self.assertRaises(ValueError):
            Task(Task.Config(use_same_classspace=True)).execute(Task.Inputs(dataset0, dataset1, class_names0, class_names1))

    def test_different_classsspace_classification_with_same_name(self):
        dataset0 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        dataset1 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        class_names0 = ['label0', 'label1', 'label2', 'label3', 'label4']
        class_names1 = ['label0', 'label1', 'label2', 'label3', 'label4']
        outputs = Task(Task.Config(use_same_classspace=False)).execute(Task.Inputs(dataset0, dataset1, class_names0, class_names1))
        self.assertEqual(len(outputs.dataset), 6)
        self.assertEqual(outputs.dataset[0], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[3], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[5], (None, torch.tensor(3)))

    def test_different_classsspace_classification(self):
        dataset0 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        dataset1 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        class_names0 = ['label0', 'label1', 'label2', 'label3', 'label4']
        class_names1 = ['label00', 'label01', 'label02', 'label03', 'label04']
        outputs = Task(Task.Config(use_same_classspace=False)).execute(Task.Inputs(dataset0, dataset1, class_names0, class_names1))
        self.assertEqual(len(outputs.dataset), 6)
        self.assertEqual(outputs.dataset[0], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[3], (None, torch.tensor(6)))
        self.assertEqual(outputs.dataset[5], (None, torch.tensor(8)))

    def test_different_classsspace_classification_with_duplicated_names(self):
        dataset0 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        dataset1 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        class_names0 = ['label0', 'label1', 'label2', 'label3', 'label4']
        class_names1 = ['label0', 'label1', 'label02', 'label03', 'label04']
        outputs = Task(Task.Config(use_same_classspace=False)).execute(Task.Inputs(dataset0, dataset1, class_names0, class_names1))
        self.assertEqual(len(outputs.dataset), 6)
        self.assertEqual(outputs.dataset[0], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[3], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[5], (None, torch.tensor(6)))

    def test_different_classsspace_object_detection_with_duplicated_names(self):
        dataset0 = FakeDataset([(None, torch.tensor([[1, 0, 0, 1, 1]])), (None, torch.tensor([[2, 0, 0, 1, 1]])), (None, torch.tensor([[3, 0, 0, 1, 1]]))])
        dataset1 = FakeDataset([(None, torch.tensor([[1, 0, 0, 1, 1]])), (None, torch.tensor([[2, 0, 0, 1, 1], [4, 0, 0, 1, 1]])), (None, torch.tensor([[3, 0, 0, 1, 1]]))])
        class_names0 = ['label0', 'label1', 'label2', 'label3', 'label4']
        class_names1 = ['label0', 'label1', 'label02', 'label03', 'label04']
        outputs = Task(Task.Config(use_same_classspace=False, task_type='object_detection')).execute(Task.Inputs(dataset0, dataset1, class_names0, class_names1))
        self.assertEqual(len(outputs.dataset), 6)
        self.assertTrue(torch.equal(outputs.dataset[0][1], torch.tensor([[1, 0, 0, 1, 1]])))
        self.assertTrue(torch.equal(outputs.dataset[3][1], torch.tensor([[1, 0, 0, 1, 1]])))
        self.assertTrue(torch.equal(outputs.dataset[5][1], torch.tensor([[6, 0, 0, 1, 1]])))

    def test_no_second_dataset(self):
        dataset0 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset0, None))
        self.assertEqual(outputs.dataset, dataset0)

    def test_get_targets(self):
        dataset0 = FakeDatasetWithGetTargets([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        dataset1 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset0, dataset1))
        self.assertEqual(len(outputs.dataset), 6)
        self.assertEqual(outputs.dataset.get_targets(0), torch.tensor(1))
        self.assertEqual(outputs.dataset.get_targets(1), torch.tensor(2))
        self.assertEqual(outputs.dataset.get_targets(2), torch.tensor(3))
        self.assertEqual(dataset0._counter, 3)
        self.assertEqual(outputs.dataset.get_targets(3), torch.tensor(1))
        self.assertEqual(outputs.dataset.get_targets(4), torch.tensor(2))
        self.assertEqual(outputs.dataset.get_targets(5), torch.tensor(3))
        self.assertEqual(dataset0._counter, 3)

    def test_phrase_grounding_dataset(self):
        dataset0 = FakeDataset([(('abc', PIL.Image.new('RGB', (10, 10))), [(0, 2), torch.tensor([[0.1, 0.1, 0.2, 0.2]])]),
                                (('def', PIL.Image.new('RGB', (10, 10))), [(0, 3), torch.tensor([[0.2, 0.2, 0.3, 0.3]])])])
        dataset1 = FakeDataset([(('abc2', PIL.Image.new('RGB', (10, 10))), [(0, 2), torch.tensor([[0.1, 0.1, 0.2, 0.2]])]),
                                (('def2', PIL.Image.new('RGB', (10, 10))), [(0, 3), torch.tensor([[0.2, 0.2, 0.3, 0.3]])])])
        outputs = Task(Task.Config(task_type='phrase_grounding')).execute(Task.Inputs(dataset0, dataset1))
        self.assertEqual(len(outputs.dataset), 4)
        self.assertEqual(outputs.dataset[0][0], ('abc', PIL.Image.new('RGB', (10, 10))))
        self.assertEqual(outputs.dataset[1][0], ('def', PIL.Image.new('RGB', (10, 10))))
        self.assertEqual(outputs.dataset[2][0], ('abc2', PIL.Image.new('RGB', (10, 10))))
        self.assertEqual(outputs.dataset[3][0], ('def2', PIL.Image.new('RGB', (10, 10))))
        self.assertTrue(torch.equal(outputs.dataset[0][1][1], torch.tensor([[0.1, 0.1, 0.2, 0.2]])))
        self.assertTrue(torch.equal(outputs.dataset[1][1][1], torch.tensor([[0.2, 0.2, 0.3, 0.3]])))
        self.assertTrue(torch.equal(outputs.dataset[2][1][1], torch.tensor([[0.1, 0.1, 0.2, 0.2]])))
        self.assertTrue(torch.equal(outputs.dataset[3][1][1], torch.tensor([[0.2, 0.2, 0.3, 0.3]])))



import base64
import dataclasses
import io
import json
import logging
import pathlib
import PIL.Image
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Save a VQA dataset to a JSONL file.

    The dataset is expected to be a torch.utils.data.Dataset with elements of the form ((question, image), answer).
    The image is expected to be a PIL.Image.Image.
    The dataset is saved as a JSONL file, where each line is a JSON object with the following fields:
    - id: the index of the element in the dataset
    - question: the question
    - image: the image, encoded as a base64 string
    - answer: the answer

    Config:
        filepath (Path): the path to save the dataset to
    """
    VERSION = '0.1.1'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Config:
        filepath: pathlib.Path

    def execute(self, inputs):
        logger.info(f"Saving dataset to {self.config.filepath}. Dataset size: {len(inputs.dataset)}")

        self.config.filepath.parent.mkdir(parents=True, exist_ok=True)
        dataset = ConvertedDataset(inputs.dataset)
        dataloader = torch.utils.data.DataLoader(dataset, batch_size=256, num_workers=8)
        with open(self.config.filepath, 'w') as file:
            for data in dataloader:
                for d in data:
                    file.write(d + '\n')

        logger.info(f"Saved dataset to {self.config.filepath}.")

        return self.Outputs()


class ConvertedDataset(torch.utils.data.Dataset):
    def __init__(self, dataset):
        self._dataset = dataset

    def __getitem__(self, index):
        (question, image), answer = self._dataset[index]
        if not isinstance(question, str) or not isinstance(answer, str):
            raise ValueError(f"Expected question and answer to be strings, got {type(question)} and {type(answer)}")
        if not isinstance(image, PIL.Image.Image):
            raise ValueError(f"Expected image to be PIL.Image.Image, got {type(image)}")

        with io.BytesIO() as f:
            image.save(f, format='jpeg')
            image_bytes = f.getvalue()

        return json.dumps({'id': index + 1, 'question': question, 'image': base64.b64encode(image_bytes).decode('ascii'), 'answer': answer})

    def __len__(self):
        return len(self._dataset)



import dataclasses
import json
import logging
import pathlib
import typing
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Extract value for each entry in a JSONL by a key.

    Config:
        key (str): key to extract value
    """
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        jsonl_file_path: pathlib.Path

    @dataclasses.dataclass
    class Outputs:
        results: typing.List[typing.Any]

    @dataclasses.dataclass
    class Config:
        key: str

    def execute(self, inputs):
        logger.info(f"Extracting values from {inputs.jsonl_file_path}, by key {self.config.key}.")

        outputs = []
        with open(inputs.jsonl_file_path, 'r') as file:
            for line in file:
                data = json.loads(line)
                outputs.append(data[self.config.key])

        return self.Outputs(results=outputs)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
from typing import Callable, Optional
import PIL.Image
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Make phrase grounding image transform.

    The input of the tramsform function is:
        ((caption, image), [[((span_start, span_end), bboxes)]]

    image_transform is applied to the image, and text_transform is applied to the caption.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        image_transform: Optional[Callable[[PIL.Image.Image], torch.Tensor]] = None
        text_transform: Optional[Callable[[str], str]] = None

    @dataclasses.dataclass
    class Outputs:
        transform: Callable

    def execute(self, inputs):
        transform = Transform(inputs.image_transform, inputs.text_transform)
        return self.Outputs(transform=transform)


class Transform:
    def __init__(self, image_transform, text_transform):
        self._image_transform = image_transform
        self._text_transform = text_transform

    def __call__(self, inputs, targets):
        caption, image = inputs
        if self._image_transform is not None:
            image = self._image_transform(image)

        if self._text_transform is not None:
            caption = self._text_transform(caption)

        return (caption, image), targets



import dataclasses
import logging
import typing
import torch
import torchvision
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Calculate precision/recall for phrase grounding.

    Config:
        top_k (Optional[int]): Number of top predictions to consider. If None, all predictions are considered. Default: 1
        iou_threshold (float): IOU threshold for positive prediction. Default: 0.5
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        predictions: typing.List[typing.List[typing.Tuple[typing.Tuple[int, int], torch.Tensor]]]
        targets: typing.List[typing.List[typing.Tuple[typing.Tuple[int, int], torch.Tensor]]]

    @dataclasses.dataclass
    class Config:
        top_k: typing.Optional[int] = 1
        iou_threshold: float = 0.5

    @dataclasses.dataclass
    class Outputs:
        precision: float = 0
        recall: float = 0

    def execute(self, inputs):
        if len(inputs.predictions) != len(inputs.targets):
            raise ValueError(f"Number of predictions ({len(inputs.predictions)}) does not match number of targets ({len(inputs.targets)})")

        num_correct = 0
        num_total_targets = 0
        num_total_predictions = 0
        for prediction, target in zip(inputs.predictions, inputs.targets):
            target_span_boxes = {tuple(span): boxes for span, boxes in target}
            assert len(target_span_boxes) == len(target), "Duplicate spans in target"
            assert len(set(tuple(pred[0]) for pred in prediction)) == len(prediction), "Duplicate spans in prediction"
            num_total_targets += len(target_span_boxes)

            for pred in prediction:
                target_boxes = target_span_boxes.get(tuple(pred[0]))
                if target_boxes is not None:
                    num_total_predictions += 1
                    if self._is_positive(pred[1], target_boxes):
                        num_correct += 1

        precision = num_correct / num_total_predictions if num_total_predictions > 0 else 0
        recall = num_correct / num_total_targets if num_total_targets > 0 else 0
        logger.info(f"Precision@{self.config.top_k} IOU({self.config.iou_threshold}): {precision}, correct: {num_correct}, total: {num_total_predictions}")
        logger.info(f"Recall@{self.config.top_k} IOU({self.config.iou_threshold}): {recall}, correct: {num_correct}, total: {num_total_targets}")

        return self.Outputs(precision=precision, recall=recall)

    def _is_positive(self, predicted_boxes, target_boxes):
        if len(predicted_boxes) == 0 or len(target_boxes) == 0:
            return False
        ious = torchvision.ops.box_iou(predicted_boxes, target_boxes)
        return ious[:self.config.top_k].max() >= self.config.iou_threshold



"""Compile the model module using torch.compile.

Args:
    backend ('inductor' or 'ipex'): The backend for torch.compile().
    dtype ('fp32' or 'bf16'): The dtype for IPEX backend.
"""
import logging
import torch
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, backend='inductor', dtype='fp32'):
        if backend not in ['inductor', 'ipex']:
            raise ValueError(f"Unsupported torch.compile backend '{backend}'")
        if dtype not in ['fp32', 'bf16']:
            raise ValueError(f"Unsupported dtype '{dtype}'")
        self._backend = backend
        self._dtype = dtype

    def on_train_start(self, trainer, model):
        logger.info(f"Compiling the model module using backend '{self._backend}'")
        self._original_model = model
        if self._backend == 'ipex':
            logger.debug("Loading intel_extension_for_pytorch")
            # Load intel_extension_for_pytorch to enable IPEX backend
            try:
                import intel_extension_for_pytorch  # noqa: F401
            except ImportError:
                logger.error("intel_extension_for_pytorch is not installed. Please install it by running 'pip install intel-extension-for-pytorch'")
                raise

            # weights_prepack is disabled since ops are well supported in ipex backend.
            dtype = {'fp32': torch.float32, 'bf16': torch.bfloat16}
            model = intel_extension_for_pytorch.optimize(model, weights_prepack=False, dtype=dtype)
            return torch.compile(model, backend=self._backend)
        else:
            return torch.compile(model, backend=self._backend, mode='max-autotune')

    def on_train_end(self, trainer, model):
        # Move the compiled model back to CPU. Otherwise, the model will stay on GPU and cause memory leak.
        model.to('cpu')

        # The OptimizedModule shares the same state_dict with the original model.
        model = self._original_model
        del self._original_model
        return model

    def on_prediction_start(self, trainer, model):
        return self.on_train_start(trainer, model)

    def on_prediction_end(self, trainer, model):
        return self.on_train_end(trainer, model)
